local easing = require("easing")
local MakePlayerCharacter = require "prefabs/player_common"

local assets = 
{
    Asset("SCRIPT", "scripts/prefabs/player_common.lua"),
}
local prefabs = 
{
	"droplets",
	"puddle_move_fx",
}

function Womp_UpdateMoisture(self, dt)
    if self.forceddry then
        return
    end

    local sleepingbagdryingrate = self:GetSleepingBagDryingRate()
    if sleepingbagdryingrate ~= nil then
        self.rate = -sleepingbagdryingrate
    else
        local moisturerate = self:GetMoistureRate()
        local dryingrate = self:GetDryingRate(moisturerate)
        local equippedmoisturerate = self:GetEquippedMoistureRate(dryingrate)

        self.rate = moisturerate + equippedmoisturerate - dryingrate
    end
	
	local h = self.inst.components.hunger:GetPercent()
	local wompRate = h > .25 and .5 or easing.linear(h * 4, 0, .5, 1) -- Womp produces less moisture over time when low on hunger.
	
	self.rate = self.rate + wompRate

    self.ratescale =
        (self.rate > .9 and RATE_SCALE.INCREASE_HIGH) or
        (self.rate > .5 and RATE_SCALE.INCREASE_MED) or
        (self.rate > .001 and RATE_SCALE.INCREASE_LOW) or
        (self.rate < -3 and RATE_SCALE.DECREASE_HIGH) or
        (self.rate < -1.5 and RATE_SCALE.DECREASE_MED) or
        (self.rate < -.001 and RATE_SCALE.DECREASE_LOW) or
        RATE_SCALE.NEUTRAL

    self:DoDelta(self.rate * dt)
end
function Womp_GetDryingRate(self, moisturerate)
	-- Don't dry if it's raining
    if (moisturerate or self:GetMoistureRate()) > 0 then
        return 0
    end

    local heaterPower = self.inst.components.temperature ~= nil and math.clamp(self.inst.components.temperature.externalheaterpower, 0, 1) or 0
	
	local t = self.inst.components.temperature

	local base = (t:IsOverheating() and 2) or -- Dries fast when overheating.
				 (self.inst.components.health.takingfiredamage and 16) or -- Dries REALLY fast when taking fire damage.
				 -4.7 + (t:GetCurrent() / t:GetMax()) -- Temperature alters drying rate a teeny bit.

    local rate = base
        + easing.linear(heaterPower, self.minPlayerTempDrying, 5, 1)
        + easing.linear(TheWorld.state.temperature, self.minDryingRate, self.maxDryingRate, self.optimalDryingTemp)
		+ easing.inExpo(self:GetMoisture(), 0, 1, self.maxmoisture)

    return math.clamp(rate, 0, self.maxDryingRate + self.maxPlayerTempDrying)
end

function Womp_SanityRecalc(self, dt)
    local total_dapperness = self.dapperness or 0
    for k, v in pairs(self.inst.components.inventory.equipslots) do
        if v.components.equippable ~= nil then
            total_dapperness = total_dapperness + v.components.equippable:GetDapperness(self.inst)
			if v.components.equippable.inst:GetIsWet() then -- Remove extra sanity loss due to having wet items equipped.
				total_dapperness = total_dapperness - TUNING.WET_ITEM_DAPPERNESS
			end
        end
    end

    total_dapperness = total_dapperness * self.dapperness_mult

    local dapper_delta = total_dapperness * TUNING.SANITY_DAPPERNESS
    
    local light_delta
    if TheWorld.state.isday and not TheWorld:HasTag("cave") then
        light_delta = -TUNING.SANITY_DAY_GAIN - (self.inst.currentRain and 0 or 0.15) --Loses sanity during the day, unless it's raining.
    else
        local lightval = CanEntitySeeInDark(self.inst) and .9 or self.inst.LightWatcher:GetLightValue()
        light_delta =
            (   (lightval > TUNING.SANITY_HIGH_LIGHT and TUNING.SANITY_NIGHT_LIGHT) or
                (lightval < TUNING.SANITY_LOW_LIGHT and TUNING.SANITY_NIGHT_DARK) or
                TUNING.SANITY_NIGHT_MID
            ) * self.night_drain_mult
    end

    local aura_delta = 0
    local x, y, z = self.inst.Transform:GetWorldPosition()
    local ents = TheSim:FindEntities(x, y, z, TUNING.SANITY_EFFECT_RANGE, nil, { "FX", "NOCLICK", "DECOR","INLIMBO" })
    for i, v in ipairs(ents) do 
        if v.components.sanityaura ~= nil and v ~= self.inst then
            local aura_val = v.components.sanityaura:GetAura(self.inst) / math.max(1, self.inst:GetDistanceSqToInst(v))
            aura_delta = aura_delta + (aura_val < 0 and aura_val * self.neg_aura_mult or aura_val)
        end
    end

    self:RecalcGhostDrain()
    local ghost_delta = TUNING.SANITY_GHOST_PLAYER_DRAIN * self.ghost_drain_mult

    self.rate = dapper_delta + light_delta + aura_delta + ghost_delta

    if self.custom_rate_fn ~= nil then
        self.rate = self.rate + self.custom_rate_fn(self.inst)
    end

    self.rate = self.rate * self.rate_modifier
    self.ratescale =
        (self.rate > .2 and RATE_SCALE.INCREASE_HIGH) or
        (self.rate > .1 and RATE_SCALE.INCREASE_MED) or
        (self.rate > .01 and RATE_SCALE.INCREASE_LOW) or
        (self.rate < -.4 and RATE_SCALE.DECREASE_HIGH) or
        (self.rate < -.15 and RATE_SCALE.DECREASE_MED) or
        (self.rate < -.02 and RATE_SCALE.DECREASE_LOW) or
        RATE_SCALE.NEUTRAL

    --print (string.format("dapper: %2.2f light: %2.2f TOTAL: %2.2f", dapper_delta, light_delta, self.rate*dt))
    self:DoDelta(self.rate * dt, true)
end
function SanityForWetness(inst)
	return (inst.wetness > 79 and inst.wetness / 3000) or -- 1.6 to 2 sanity per minute while at 80 wetness or higher.
			(inst.wetness < 6 and -1) or -- -1 sanity per second when at 5 wetness or below
			(inst.wetness < 21 and -0.5) or -- -.5 sanity per second when at 20 wetness or below (but still above 5)
			0 --No sanity change when between 20 and 80 wetness.
end

local function dorainstuff(inst, dt) --Gain health at random ticks while raining
	if inst.components.moisture ~= nil and inst.components.moisture:GetMoisture() > 0 then
		local t = GetTime()
		
		if inst.components.inventory:GetEquippedMoistureRate(EQUIPSLOTS.HEAD) <= 0 and inst.components.moisture:GetRate() > 0 then
			local waterproofmult =
                (   inst.components.sheltered ~= nil and
                    inst.components.sheltered.sheltered and
                    inst.components.sheltered.waterproofness or 0
                ) +
                (   inst.components.inventory ~= nil and
                    inst.components.inventory:GetWaterproofness() or 0
                )
			if waterproofmult < 1 and t > inst.rt_time + inst.rt_time_offset + waterproofmult * 7 then --Does not heal in rain when waterproof.
				inst.components.health:DoDelta(3, false, "Rain")
				inst.rt_time_offset = 5 + math.random() * 3
				inst.rt_time = t
			end
		end
	end
end

local function DoDryStuff(inst, dt) -- Takes random ticks of damage when too dry.
	local t = GetTime()
	
	if t > inst.dry_time + inst.dry_time_offset then
		local damage = easing.outInCirc((inst.wetness / 20), 3, 1 - 3, 1)
		inst.components.health:DoDelta(-damage, false, "Dryness")
		inst.dry_time_offset = 1 + math.random() * 3
		inst.dry_time = t
	end
end

local function GetSpeed(inst) --Moves faster when wetness is high, and even faster when raining.
	local speed = (inst.wetness > 79 and 1.1) or
		   (inst.wetness > 39 and 1) or
		   (inst.wetness > 19 and 0.95) or
		   (inst.wetness > 4 and 0.85) or
		   0.75
		   
    return inst.currentRain and speed + 0.15 or speed
end

local function Womp_Onhit(inst, attacker, damage)
	local dryby = damage / 4
	
	inst.components.moisture:DoDelta(-dryby)
	
	local x, y, z = inst.Transform:GetWorldPosition()
	y = y + .25 + math.random() * 2
	SpawnPrefab("droplets").Transform:SetPosition(x, y, z)
	if attacker ~= nil and attacker.components.moisture ~= nil then -- Attackers gain wetness when hitting Womp, really only a pvp thing I guess.
		attacker.components.moisture:DoDelta(dryby / 2)
	end
end

local function SpawnMoveFx(inst)
    if math.random() < .4 then
        local fx = SpawnPrefab("puddle_move_fx")
        if fx ~= nil then
            if inst.sg.mem.recentfx == nil then
                inst.sg.mem.recentfx = {}
            end
            local recentcount = #inst.sg.mem.recentfx
            local rand = math.random(7 - recentcount)
            if recentcount > 0 then
                while table.contains(inst.sg.mem.recentfx, rand) do
                    rand = rand + 1
                end
                if recentcount >= 6 then
                    table.remove(inst.sg.mem.recentfx, 1)
                end
            end
            table.insert(inst.sg.mem.recentfx, rand)
            local basescale = inst.Transform:GetScale()
            local randscale = math.random() * math.sqrt(1.3 - .5)
            fx:SetVariation(rand, basescale * (1.3 - randscale * randscale))
            fx.Transform:SetPosition(inst.Transform:GetWorldPosition())
        end
    end
    inst.sg.mem.lastspawnmovefx = GetTime()
end

local function Womp_Update(inst)
	inst.wetness = inst.components.moisture:GetMoisture()
	
	if inst.wetness > 19 and inst.sg:HasStateTag("moving") then -- Neat puddle effect.
		SpawnMoveFx(inst)
	end
	
	if inst.wetness < 20 and inst.dry_task == nil then
		inst.dry_task = inst:DoPeriodicTask(.1, DoDryStuff, nil, .1)
	elseif inst.dry_task ~= nil then
		inst.dry_task:Cancel()
		inst.dry_task = nil
	end
	
	inst.components.locomotor:SetExternalSpeedMultiplier(inst, "womp_speed_mod", GetSpeed(inst))
	inst.components.temperature.inherentsummerinsulation = inst.wetness / 2 --Gains half of his wetness as summer insulation - stays cool.
	inst.components.temperature.inherentinsulation = inst.wetness / -4 --Loses a quarter of his wetness as winter insulation - gets cold. 
	inst.components.health.fire_damage_scale = -1 * (inst.wetness / 100) / 1 + 1.5 --Takes half fire damage at max wetness, 1.5x fire damage at 0 wetness, linear scaling in between.
end

local function onisraining(inst, israining)
	if israining then
		inst.currentRain = true
		if inst.rain_task == nil then
			inst.rain_task = inst:DoPeriodicTask(.1, dorainstuff, nil, .1)
		end
	else
		inst.currentRain = false
		if inst.rain_task ~= nil then
			inst.rain_task:Cancel()
			inst.rain_task = nil
		end
	end
end

-- When the character is revived
local function onbecamehuman(inst)
	if inst.keepWet == nil then
		inst.keepWet = inst:DoPeriodicTask(.1, Womp_Update, nil, .1)
	end
	
	if not inst.watchingrain then
		inst.watchingrain = true
		inst:WatchWorldState("israining", onisraining)
		onisraining(inst, TheWorld.state.israining)
	end
end

local function onbecameghost(inst)
	-- Remove speed modifier when becoming a ghost
	inst.components.locomotor:RemoveExternalSpeedMultiplier(inst, "womp_speed_mod")
   
	if inst.keepWet ~= nil then
		inst.keepWet:Cancel()
		inst.keepWet = nil
	end
   
    if inst.dry_task ~= nil then
		inst.dry_task:Cancel()
		inst.dry_task = nil
	end
   
	if inst.rain_task ~= nil then
		inst.rain_task:Cancel()
		inst.rain_task = nil
	end
	
	if inst.watchingrain then
		inst.watchingrain = false
		inst:StopWatchingWorldState("israining", onisraining)
	end
end

-- When loading or spawning the character
local function onload(inst)
	inst:ListenForEvent("ms_respawnedfromghost", function(inst) inst.components.moisture:SetMoistureLevel(85) end)
    inst:ListenForEvent("ms_respawnedfromghost", onbecamehuman)
    inst:ListenForEvent("ms_becameghost", onbecameghost)

    if inst:HasTag("playerghost") then
        onbecameghost(inst)
    else
        onbecamehuman(inst)
    end
end

-- This initializes for both the server and client. Tags can be added here.
local common_postinit = function(inst)

	inst:AddTag("swampbro")
	inst:AddTag("wet")
	-- Womp is a monster lore wise, if you're wondering. He's not tagged as one because getting spiders to be hostile to monsters is more annoying than getting the neutrals to be hostile.
	--inst:AddTag("monster") 

	-- Minimap icon
	inst.MiniMapEntity:SetIcon("womp.tex")
	
	inst.foleysound = "womp/characters/womp/foleysound"
end

-- This initializes for the server only. Components are added here.
local master_postinit = function(inst)
	
	--Stuff for staying wet / drying
	inst.wetness = inst.components.moisture:GetMoisture()
	inst.keepWet = nil
	inst.dry_task = nil
	inst.dry_time = 0
	inst.dry_time_offset = 1
	inst.components.temperature.maxmoisturepenalty = 0 -- Doesn't get extra cold because of wetness - he's used to it.
	inst.components.moisture.maxMoistureRate = 4
	inst.components.moisture.maxPlayerTempDrying = 20 -- Increase the max drying rate because fire damage should make you dry REALLY fast.
	inst.components.moisture.OnUpdate = Womp_UpdateMoisture
	inst.components.moisture.GetDryingRate = Womp_GetDryingRate
	inst.components.moisture.GetSleepingBagDryingRate = function() return nil end
	
	--Stuff for rain
	inst.rain_task = nil
	inst.rt_time = 0
	inst.rt_time_offset = 5
	inst.watchingrain = false
	inst.currentRain = false
	
	--Prevent tool slip during onwork - Thanks DarkXero
	local listeners1 = inst.event_listeners.working
	local listener_fns1 = listeners1[inst]
	local old1 = listener_fns1[1]
	inst:RemoveEventCallback("working", old1)
	--Prevent tool slip when attacking
	local listeners2 = inst.event_listeners.onattackother
	local listener_fns2 = listeners2[inst]
	local old2 = listener_fns2[1]
	listener_fns2[1] = function(inst, data)
		data.weapon = nil
		return old2(inst, data)
	end

	-- choose which sounds this character will play
	inst.soundsname = "womp"
	
	inst.components.eater.strongstomach = true
	inst.components.eater.ignoresspoilage = true
	
	-- Stats	
	inst.components.health:SetMaxHealth(120) -- Low HP to balance the dodge chance, Womp is squishy after all.
	inst.components.hunger:SetMax(200)		 -- High hunger cause he didn't eat often in the abyss.
	inst.components.sanity:SetMax(100)		 -- Low sanity cause he is a monster.
	
	inst.components.sanity.night_drain_mult = 0 -- Womp is used to darkness in the void.
	inst.components.sanity.Recalc = Womp_SanityRecalc
	inst.components.sanity.custom_rate_fn = SanityForWetness
	
    inst.components.combat.onhitfn = Womp_Onhit
	inst._GetAttacked = inst.components.combat.GetAttacked
	inst.components.combat.GetAttacked = function(attacker, damage, weapon, stimuli) -- Dodge chance when 20 wetness or above.
		if inst.wetness > 19 and math.random() > (inst.wetness / 100) / 2 then
			inst._GetAttacked(attacker, damage, weapon, stimuli)
		else
			inst.SoundEmitter:PlaySound("womp/characters/womp/whoosh")
		end
	end
	
	inst.OnLoad = onload
    inst.OnNewSpawn = function(inst) 
		onload(inst) 
		inst.components.moisture:SetMoistureLevel(85) 
	end
	
end

return MakePlayerCharacter("womp", prefabs, assets, common_postinit, master_postinit)
